from collections.abc import Sequence
from pathlib import Path
from typing import Any

import blacksmith
from blacksmith.domain.registry import HttpResource, registry
from jinja2 import Environment, PackageLoader
from pydantic import BaseModel, Field

env = Environment(loader=PackageLoader("whitesmith"), autoescape=False)
redoc_template = env.get_template("redoc.jinja2")

sd = blacksmith.SyncRouterDiscovery(
    service_url_fmt="http://{service}.{version}",
    unversioned_service_url_fmt="http://{service}.NaN",
)


JSONSchema = dict[str, Any]


class Parameter(BaseModel):
    name: str
    in_: str = Field(alias="in")  # path, query, header
    required: bool = Field(default=False)
    schema_: JSONSchema | None = Field(alias="schema", default=None)


class MediaType(BaseModel):
    schema_: JSONSchema | None = Field(alias="schema", default=None)


class Content(BaseModel):
    application_json: MediaType = Field(default=None, alias="application/json")


class Response(BaseModel):
    description: str = Field(default="OK")
    content: Content | None = Field(default=None)


class Operation(BaseModel):
    operationId: str
    parameters: list[Parameter] = Field(default_factory=list)
    responses: dict[str, Response]
    summary: str
    tags: list[str]


class Components(BaseModel):
    schemas: dict[str, JSONSchema] = Field(default_factory=dict)


class Info(BaseModel):
    title: str
    version: str


class Server(BaseModel):
    url: str
    description: str | None = None


class OpenAPIDocument(BaseModel):
    openapi: str = "3.1.0"
    info: Info
    paths: dict[str, dict[str, Operation]] = Field(default_factory=dict)
    components: Components = Field(default_factory=Components)
    servers: list[Server] = Field(default_factory=list)


def request_schema_to_parameter(request: type[blacksmith.Request]) -> list[Parameter]:
    params: list[Parameter] = []
    json_schemas = request.model_json_schema()
    for name, field in request.model_fields.items():
        param = Parameter.model_validate(
            {
                "name": field.alias or name,
                "in": field.json_schema_extra["location"],  # type: ignore
                # "required": field.is_required,
                "schema": json_schemas["properties"][name],
            }
        )
        params.append(param)
    return params


def response_schema_to_responses(
    response: type[blacksmith.Response] | None,
) -> tuple[dict[str, Response], dict[str, JSONSchema]]:
    schemas: dict[str, JSONSchema] = {}
    responses: dict[str, Response] = {}
    if response is None:
        responses["204"] = Response(description="No Content")
    else:
        schema_name = f"{response.__module__}__{response.__qualname__}"
        schema_name = schema_name.replace(".", "_")
        responses["200"] = Response(
            description=response.__doc__ or response.__name__,
            content=Content.model_validate(
                {
                    # we should read something for the accepted content type
                    "application/json": MediaType(
                        schema={"$ref": f"#/components/schemas/{schema_name}"}
                    ),
                }
            ),
        )
        schemas[schema_name] = response.model_json_schema()

    return responses, schemas


def add_resource(
    openapi: OpenAPIDocument,
    resource: HttpResource,
    client: str,
    resource_name: str,
    summary: str,
    tags: list[str],
) -> None:
    if resource.contract:
        for http_method, schemas in resource.contract.items():
            method = http_method.lower()
            responses, resp_schemas = response_schema_to_responses(schemas[1])
            openapi.components.schemas.update(resp_schemas)
            if resource.path not in openapi.paths:
                openapi.paths[resource.path] = {}
            openapi.paths[resource.path][method] = Operation(
                operationId=f"{client}_{method}_{resource_name}",
                summary=summary,
                parameters=request_schema_to_parameter(schemas[0]),
                responses=responses,
                tags=tags,
            )


def generate_openapis(
    outdir: str | Path, resources_mod: Sequence[str], overwrite: bool
) -> None:
    outdir = Path(outdir)
    outdir.mkdir(parents=True, exist_ok=True)
    blacksmith.scan(*resources_mod)

    print("Generating openapis from blacksmith registry...")

    services = []
    for client, service in registry.client_service.items():
        service, resources = registry.get_service(client)
        endpoint = sd.get_endpoint(*service)

        print(f"Processing client {client}...")
        openapi = OpenAPIDocument(
            info=Info(
                title=f"{client} for service {service[0]}",
                version=service[1] or "unversionned",
            ),
            servers=[
                Server(
                    url=endpoint,
                    description="Client Contract Testing",
                )
            ],
        )
        for name, resource in resources.items():
            print(f"Prosessing {name}...")
            if resource.collection:
                add_resource(
                    openapi,
                    resource.collection,
                    client,
                    f"{name}_collection",
                    summary=f"Collection or {name}",
                    tags=[name],
                )

            if resource.resource:
                add_resource(
                    openapi, resource.resource, client, name, summary=name, tags=[name]
                )

        file_ = outdir / f"{client}.json"
        services.append(
            {
                "file": f"/{client}.json",
                "name": client,
                "id": client.replace(" ", "-"),
                "endpoint": endpoint,
            }
        )
        if overwrite or not file_.exists():
            print(f"Writing {file_}")
            file_.write_text(openapi.model_dump_json(by_alias=True, exclude_none=True))

    file_ = outdir / "index.html"
    if overwrite or not file_.exists():
        print(f"Writing {file_}")
        redoc = redoc_template.render(
            services=services,
        )
        file_.write_text(redoc)



from collections.abc import Sequence
from pathlib import Path

import blacksmith
from blacksmith.domain.registry import HttpResource, registry
from jinja2 import Environment, PackageLoader
from pydantic import BaseModel, Field

env = Environment(loader=PackageLoader("whitesmith"), autoescape=False)
handlers_template = env.get_template("handlers.jinja2")

sd = blacksmith.SyncRouterDiscovery(
    service_url_fmt="http://{service}.{version}",
    unversioned_service_url_fmt="http://{service}.NaN",
)


def import_pydantic_factory() -> str:
    return "from polyfactory.factories.pydantic_factory import ModelFactory"


class ResponseModel(BaseModel):
    mod: str = Field(...)
    name: str = Field(...)

    def __hash__(self) -> int:
        return hash(f"{self.mod}:{self.name}")


class Route(BaseModel):
    python_method: str = Field(...)
    http_method: str = Field(...)
    url_pattern: str = Field(...)
    response_model: str = Field(...)
    response_class: str = Field(...)


class HandlerTemplateContext(BaseModel):
    extra_import_lines: set[str] = Field(default_factory=set)
    whitesmith_imports: set[str] = Field(default_factory=lambda: {"router"})
    has_missing_schema: bool = Field(default=False)
    response_models: set[ResponseModel] = Field(default_factory=set)
    routes: list[Route] = Field(default_factory=list)

    @property
    def sorted_response_models(self) -> list[ResponseModel]:
        return sorted(self.response_models, key=lambda x: x.__class__.__name__)

    def add_resource(
        self,
        endpoint: str,
        service: str,
        name: str,
        resource: HttpResource | None,
        prefix: str = "",
    ) -> "HandlerTemplateContext":
        if not resource or not resource.contract:
            return self
        for method, schemas in resource.contract.items():
            response_schema = schemas[1]
            response_schema_name = ""
            if response_schema:
                response_schema_name = response_schema.__name__
                self.extra_import_lines.add(import_pydantic_factory())
                self.response_models.add(
                    ResponseModel(
                        mod=response_schema.__module__,
                        name=response_schema.__name__,
                    )
                )
            else:
                self.has_missing_schema = True

            response_class = "HTTPResponse"
            if prefix == "collection_" and method == "GET":
                response_class = "HTTPCollectionResponse"
            self.whitesmith_imports.add(response_class)
            self.routes.append(
                Route(
                    python_method=f"{service}_{name}_{prefix}{method}".lower(),
                    http_method=f"{method}",
                    url_pattern=f"{endpoint}{resource.path}",
                    response_model=response_schema_name,
                    response_class=response_class,
                )
            )
        return self


def generate_handlers(
    outdir: str | Path, resources_mod: Sequence[str], overwrite: bool
) -> None:
    outdir = Path(outdir)
    blacksmith.scan(*resources_mod)

    print("Generating mocks from blacksmith registry...")

    outdir = outdir / "whitesmith_handlers"
    outdir.mkdir(parents=True, exist_ok=True)
    for client, service in registry.client_service.items():
        service, resources = registry.get_service(client)
        endpoint = sd.get_endpoint(*service)

        print(f"Processing client {client}...")

        context = HandlerTemplateContext()
        for name, resource in resources.items():
            context = context.add_resource(
                endpoint, client, name, resource.collection, prefix="collection_"
            ).add_resource(endpoint, client, name, resource.resource)

        file_ = outdir / "__init__.py"
        if overwrite or not file_.exists():
            print(f"Writing {file_}")
            file_.write_text("")

        file_ = outdir / f"{client}.py"
        if overwrite or not file_.exists():
            print(f"Writing {file_}")
            handler = handlers_template.render(context=context)
            file_.write_text(handler)



from blacksmith import (
    AsyncAbstractTransport,
    HTTPError,
    HTTPRequest,
    HTTPResponse,
    HTTPTimeout,
    SyncAbstractTransport,
)

from .router import Router


class SyncFakeTransport(SyncAbstractTransport):
    def __init__(self, router: Router):
        super().__init__()
        self.router = router

    def __call__(
        self,
        req: HTTPRequest,
        client_name: str,
        path: str,
        timeout: HTTPTimeout,
    ) -> HTTPResponse:
        """Process the HTTP request."""
        key = f"{req.method} {req.url.format(**req.path)}"

        resp = self.router.handle(req)

        if resp.status_code >= 400:
            raise HTTPError(
                f"{resp.status_code} FAKE FAIL {key}",
                req,
                resp,
            )

        return resp


class AsyncFakeTransport(AsyncAbstractTransport):
    def __init__(self, router: Router) -> None:
        super().__init__()
        self._fake_transport = SyncFakeTransport(router)

    async def __call__(
        self,
        req: HTTPRequest,
        client_name: str,
        path: str,
        timeout: HTTPTimeout,
    ) -> HTTPResponse:
        """Process the HTTP request."""
        ret = self._fake_transport(req, client_name, path, timeout)
        return ret



from collections.abc import Callable, MutableMapping
from types import ModuleType
from typing import Any

import venusian
from blacksmith import HTTPRequest
from blacksmith import HTTPResponse as BMResponse
from blacksmith.typing import HTTPMethod

from whitesmith import HTTPCollectionResponse, HTTPResponse

Handler = Callable[[HTTPRequest], HTTPResponse[Any] | HTTPCollectionResponse[Any]]


VENUSIAN_CATEGORY = "whitesmith"


class Router:
    def __init__(self) -> None:
        self.routes: MutableMapping[tuple[str, str], Handler] = {}

    def handle(self, req: HTTPRequest) -> BMResponse:
        try:
            return self.routes[(req.method, req.url_pattern)](req).build()
        except KeyError:
            return BMResponse(
                500,
                {},
                {
                    "message": f"Unregister whitesmith route: "
                    f"{req.method} {req.url_pattern}"
                },
            )

    def register(self, meth: HTTPMethod, url: str, fn: Handler) -> None:
        self.routes[(meth, url)] = fn


class RouterBuilder:
    def __init__(self) -> None:
        self.scanned: set[ModuleType] = set()

    def get(self, url: str) -> Callable[[Handler], Handler]:
        def configure(wrapped: Handler) -> Handler:
            def callback(scanner: venusian.Scanner, name: str, ob: Handler) -> None:
                if not hasattr(scanner, "router"):
                    return  # coverage: ignore
                scanner.router.register("GET", url, wrapped)  # type: ignore

            venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
            return wrapped

        return configure

    def post(self, url: str) -> Callable[[Handler], Handler]:
        def configure(wrapped: Handler) -> Handler:
            def callback(scanner: venusian.Scanner, name: str, ob: Handler) -> None:
                if not hasattr(scanner, "router"):
                    return  # coverage: ignore
                scanner.router.register("POST", url, wrapped)  # type: ignore

            venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
            return wrapped

        return configure

    def put(self, url: str) -> Callable[[Handler], Handler]:
        def configure(wrapped: Handler) -> Handler:
            def callback(scanner: venusian.Scanner, name: str, ob: Handler) -> None:
                if not hasattr(scanner, "router"):
                    return  # coverage: ignore
                scanner.router.register("PUT", url, wrapped)  # type: ignore

            venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
            return wrapped

        return configure

    def patch(self, url: str) -> Callable[[Handler], Handler]:
        def configure(wrapped: Handler) -> Handler:
            def callback(scanner: venusian.Scanner, name: str, ob: Handler) -> None:
                if not hasattr(scanner, "router"):
                    return  # coverage: ignore
                scanner.router.register("PATCH", url, wrapped)  # type: ignore

            venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
            return wrapped

        return configure

    def delete(self, url: str) -> Callable[[Handler], Handler]:
        def configure(wrapped: Handler) -> Handler:
            def callback(scanner: venusian.Scanner, name: str, ob: Handler) -> None:
                if not hasattr(scanner, "router"):
                    return  # coverage: ignore
                scanner.router.register("DELETE", url, wrapped)  # type: ignore

            venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
            return wrapped

        return configure

    def options(self, url: str) -> Callable[[Handler], Handler]:
        def configure(wrapped: Handler) -> Handler:
            def callback(scanner: venusian.Scanner, name: str, ob: Handler) -> None:
                if not hasattr(scanner, "router"):
                    return  # coverage: ignore
                scanner.router.register("OPTIONS", url, wrapped)  # type: ignore

            venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
            return wrapped

        return configure

    def head(self, url: str) -> Callable[[Handler], Handler]:
        def configure(wrapped: Handler) -> Handler:
            def callback(scanner: venusian.Scanner, name: str, ob: Handler) -> None:
                if not hasattr(scanner, "router"):
                    return  # coverage: ignore
                scanner.router.register("HEAD", url, wrapped)  # type: ignore

            venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
            return wrapped

        return configure

    def build_router(self, mod: ModuleType) -> Router:
        assert mod
        router = Router()
        scanner = venusian.Scanner(router=router)
        scanner.scan(mod, categories=[VENUSIAN_CATEGORY])
        return router


router = RouterBuilder()



from importlib import metadata

from .model import HTTPCollectionResponse, HTTPResponse
from .router import Router, router

__version__ = metadata.version("whitesmith")
__all__ = ["HTTPResponse", "HTTPCollectionResponse", "Router", "router", "__version__"]
